@@ -10,7 +10,7 @@ use tokio::{
     time::timeout,
 };
 
-use crate::prelude::ExtensionProxy;
+// use crate::prelude::ExtensionProxy;
 
 pub type ExtensionResultSender<T> = Sender<ExtensionResult<T>>;
 
@@ -57,7 +57,8 @@ impl ExtensionBus {
         contents: &Bytes,
     ) -> Result<(), Box<dyn std::error::Error>> {
         let path = self.path.join(name).with_extension("tanoshi");
-        ExtensionProxy::compile(contents, &path)?;
+        // ExtensionProxy::compile(contents, &path)?;
+        tokio::fs::write(&path, contents).await?;
 
         Ok(self.tx.send(Command::Load(
             path.to_str().ok_or("path can't to string")?.to_string(),
@@ -65,7 +66,14 @@ impl ExtensionBus {
     }
 
     pub async fn unload(&self, source_id: i64) -> Result<(), Box<dyn std::error::Error>> {
-        Ok(self.tx.send(Command::Unload(source_id))?)
+        let detail = self.detail(source_id).await?;
+
+        self.tx.send(Command::Unload(source_id))?;
+
+        let path = self.path.join(detail.name).with_extension("tanoshi");
+        info!("removing {}", path.display());
+        tokio::fs::remove_file(&path).await?;
+        Ok(())
     }
 
     pub async fn exist(&self, source_id: i64) -> Result<bool, Box<dyn std::error::Error>> {
