@@ -1,11 +1,14 @@
 use serde::{de::DeserializeOwned, Serialize};
-use std::{collections::HashMap, fmt::Debug, path::Path, sync::Arc};
+use std::{collections::BTreeMap, fmt::Debug, path::Path, sync::Arc};
 use tanoshi_lib::prelude::{Chapter, Extension, ExtensionResult, Filters, Manga, Param, Source};
 use tanoshi_util::http::Request;
 use tokio::{
     sync::mpsc::{UnboundedReceiver, UnboundedSender},
     task::JoinHandle,
+    time::Instant,
 };
+#[cfg(feature = "compiler")]
+use wasmer::Target;
 use wasmer::{imports, ChainableNamedResolver, Function, Instance, Module, Store, WasmerEnv};
 
 use wasmer_wasi::{Pipe, WasiEnv, WasiState};
@@ -27,7 +30,7 @@ impl ExtensionProxy {
         store: &Store,
         path: P,
     ) -> Result<Arc<dyn Extension>, Box<dyn std::error::Error>> {
-        let module = unsafe { Module::deserialize_from_file(&store, path)? };
+        let module = unsafe { Module::deserialize_from_file(store, path)? };
 
         let stdin = Pipe::new();
         let stdout = Pipe::new();
@@ -45,12 +48,12 @@ impl ExtensionProxy {
 
         let tanoshi = imports! {
             "tanoshi" => {
-                "host_http_request" => Function::new_native_with_env(&store, env.clone(), host_http_request),
-                "host_debug" => Function::new_native_with_env(&store, env.clone(), host_debug),
-                "host_error" => Function::new_native_with_env(&store, env.clone(), host_error),
-                "host_info" => Function::new_native_with_env(&store, env.clone(), host_info),
-                "host_trace" => Function::new_native_with_env(&store, env.clone(), host_trace),
-                "host_warn" => Function::new_native_with_env(&store, env.clone(), host_warn),
+                "host_http_request" => Function::new_native_with_env(store, env.clone(), host_http_request),
+                "host_debug" => Function::new_native_with_env(store, env.clone(), host_debug),
+                "host_error" => Function::new_native_with_env(store, env.clone(), host_error),
+                "host_info" => Function::new_native_with_env(store, env.clone(), host_info),
+                "host_trace" => Function::new_native_with_env(store, env.clone(), host_trace),
+                "host_warn" => Function::new_native_with_env(store, env.clone(), host_warn),
             }
         };
 
@@ -59,57 +62,53 @@ impl ExtensionProxy {
         Ok(Arc::new(ExtensionProxy { instance, env }))
     }
 
-    #[cfg(not(feature = "disable-compiler"))]
-    fn init_store() -> Store {
-        #[cfg(feature = "cranelift")]
-        let compiler = wasmer_compiler_cranelift::Cranelift::new();
-        #[cfg(all(feature = "llvm", not(feature = "cranelift")))]
+    #[cfg(feature = "compiler")]
+    fn init_store(target: Target) -> Store {
         let compiler = wasmer_compiler_llvm::LLVM::new();
 
-        #[cfg(feature = "universal")]
-        let engine = wasmer::Universal::new(compiler).engine();
-        #[cfg(all(feature = "dylib", not(feature = "universal")))]
-        let engine = wasmer_engine_dylib::Dylib::new(compiler).engine();
+        let engine = wasmer_engine_dylib::Dylib::new(compiler)
+            .target(target)
+            .engine();
 
         Store::new(&engine)
     }
 
     fn init_store_headless() -> Store {
-        #[cfg(feature = "dylib")]
         let engine = wasmer_engine_dylib::Dylib::headless().engine();
-        #[cfg(feature = "universal")]
-        let engine = wasmer::Universal::headless().engine();
 
         Store::new(&engine)
     }
 
-    #[cfg(not(feature = "disable-compiler"))]
-    pub fn compile_from_file<P: AsRef<Path>>(path: P) -> Result<(), Box<dyn std::error::Error>> {
+    #[cfg(feature = "compiler")]
+    pub fn compile_from_file<P: AsRef<Path>>(
+        path: P,
+        target: Target,
+    ) -> Result<(), Box<dyn std::error::Error>> {
         let wasm_bytes = std::fs::read(&path)?;
 
         let output = std::path::PathBuf::new()
             .join(&path)
-            .with_extension("tanoshi");
-        Self::compile(&wasm_bytes, output)?;
+            .with_extension(format!("{}.tanoshi", target.triple()));
+        Self::compile(&wasm_bytes, output, target)?;
 
-        std::fs::remove_file(path)?;
         Ok(())
     }
 
-    #[cfg(not(feature = "disable-compiler"))]
+    #[cfg(feature = "compiler")]
     pub fn compile<P: AsRef<Path>>(
         wasm_bytes: &[u8],
         output: P,
+        target: Target,
     ) -> Result<(), Box<dyn std::error::Error>> {
-        let store = Self::init_store();
+        let store = Self::init_store(target);
 
         debug!("compiling extension");
         let module = Module::new(&store, wasm_bytes)?;
         debug!("done");
 
-        debug!("remove wasm file");
         Ok(module.serialize_to_file(output)?)
     }
+
     fn call<T>(&self, name: &str) -> Result<T, Box<dyn std::error::Error>>
     where
         T: DeserializeOwned,
@@ -198,7 +197,7 @@ pub async fn load<P: AsRef<Path>>(
         }
     }
 
-    #[cfg(not(feature = "disable-compiler"))]
+    #[cfg(feature = "compiler")]
     compile(&path).await?;
 
     for entry in std::fs::read_dir(&path)?
@@ -220,8 +219,28 @@ pub async fn load<P: AsRef<Path>>(
     Ok(())
 }
 
-#[cfg(not(feature = "disable-compiler"))]
+#[cfg(feature = "compiler")]
 pub async fn compile<P: AsRef<Path>>(path: P) -> Result<(), Box<dyn std::error::Error>> {
+    compile_with_target(path, env!("TARGET"), true).await?;
+
+    Ok(())
+}
+
+#[cfg(feature = "compiler")]
+pub async fn compile_with_target<P: AsRef<Path>>(
+    path: P,
+    triple: &str,
+    remove_wasm: bool,
+) -> Result<(), Box<dyn std::error::Error>> {
+    use std::str::FromStr;
+    use wasmer::{CpuFeature, RuntimeError, Triple};
+
+    info!("compile wasm for {}", triple);
+
+    let triple = Triple::from_str(triple).map_err(|error| RuntimeError::new(error.to_string()))?;
+    let cpu_feature = CpuFeature::set();
+    let target = Target::new(triple, cpu_feature);
+
     match std::fs::read_dir(&path) {
         Ok(_) => {}
         Err(_) => {
@@ -236,7 +255,12 @@ pub async fn compile<P: AsRef<Path>>(path: P) -> Result<(), Box<dyn std::error::
     {
         let path = entry.path();
         info!("found wasm file at {:?}", path.clone());
-        ExtensionProxy::compile_from_file(path).map_err(|e| format!("{}", e))?;
+        ExtensionProxy::compile_from_file(&path, target.clone()).map_err(|e| format!("{}", e))?;
+
+        if remove_wasm {
+            debug!("remove wasm file");
+            std::fs::remove_file(path)?;
+        }
     }
 
     Ok(())
@@ -244,7 +268,7 @@ pub async fn compile<P: AsRef<Path>>(path: P) -> Result<(), Box<dyn std::error::
 
 async fn thread(extension_receiver: UnboundedReceiver<Command>) {
     let mut recv = extension_receiver;
-    let mut extension_map: HashMap<i64, Arc<dyn Extension>> = HashMap::new();
+    let mut extension_map: BTreeMap<i64, (Source, Arc<dyn Extension>)> = BTreeMap::new();
 
     let store = ExtensionProxy::init_store_headless();
 
@@ -253,20 +277,27 @@ async fn thread(extension_receiver: UnboundedReceiver<Command>) {
         if let Some(cmd) = cmd {
             match cmd {
                 Command::Insert(source_id, proxy) => {
-                    extension_map.insert(source_id, proxy);
+                    let source = proxy.detail();
+                    extension_map.insert(source_id, (source, proxy));
                 }
                 Command::Load(path) => {
-                    info!("load plugin from {:?}", path.clone());
-                    match ExtensionProxy::load(&store, path) {
+                    info!("load plugin from {:?}", path);
+                    let now = Instant::now();
+                    match ExtensionProxy::load(&store, &path) {
                         Ok(proxy) => {
                             let source = proxy.detail();
-                            info!("loaded: {:?}", source);
-                            extension_map.insert(source.id, proxy);
+                            info!("loaded in {} ms: {:?}", now.elapsed().as_millis(), source);
+                            extension_map.insert(source.id, (source, proxy));
                         }
                         Err(e) => {
                             error!("error load extension: {}", e);
                         }
                     }
+                    info!(
+                        "Extension {} loaded in {:?} ms",
+                        path,
+                        now.elapsed().as_millis()
+                    );
                 }
                 Command::Unload(source_id) => {
                     extension_map.remove(&source_id);
@@ -274,24 +305,24 @@ async fn thread(extension_receiver: UnboundedReceiver<Command>) {
                 Command::Exist(source_id, tx) => {
                     let exist = extension_map.get(&source_id).is_some();
                     if tx.send(exist).is_err() {
-                        error!("receiver dropped");
+                        error!("[Command::Exist] receiver dropped");
                     }
                 }
                 Command::List(tx) => {
                     let sources = extension_map
-                        .iter()
-                        .map(|(_, ext)| ext.detail())
+                        .values()
+                        .cloned()
+                        .map(|(source, _)| source)
                         .collect::<Vec<Source>>();
 
                     if tx.send(sources).is_err() {
-                        error!("receiver dropped");
+                        error!("[Command::List] receiver dropped");
                     }
                 }
                 Command::Detail(source_id, tx) => match extension_map.get(&source_id) {
-                    Some(proxy) => {
-                        let res = proxy.detail();
-                        if tx.send(res).is_err() {
-                            error!("receiver dropped");
+                    Some((detail, _)) => {
+                        if tx.send(detail.clone()).is_err() {
+                            error!("[Command::Detail] receiver dropped");
                         }
                     }
                     None => {
@@ -325,7 +356,7 @@ async fn thread(extension_receiver: UnboundedReceiver<Command>) {
 }
 
 fn process<F, T>(
-    extension_map: &HashMap<i64, Arc<dyn Extension>>,
+    extension_map: &BTreeMap<i64, (Source, Arc<dyn Extension>)>,
     source_id: i64,
     tx: ExtensionResultSender<T>,
     f: F,
@@ -335,13 +366,11 @@ fn process<F, T>(
 {
     match extension_map.get(&source_id) {
         Some(proxy) => {
-            let proxy = proxy.clone();
-            tokio::spawn(async move {
-                let res = f(proxy);
-                if tx.send(res).is_err() {
-                    error!("receiver dropped");
-                }
-            });
+            let (_, proxy) = proxy.clone();
+            let res = f(proxy);
+            if tx.send(res).is_err() {
+                error!("[process] receiver dropped");
+            }
         }
         None => {
             error!("extension with id {} not found", source_id);
